#include <avr/io.h>
#include <avr/interrupt.h>

#include <util/delay.h>

#include <stdbool.h>
#include <stdio.h>
#include <string.h>

#include "lcd.h"
#include "pff.h"
#include "light_ws2812.h"

#define MAXPIX 56
#define COLORLENGTH (MAXPIX/2)
#define FADE (256/COLORLENGTH)

struct cRGB colors[8];
struct cRGB led[MAXPIX];

FATFS fs;					// sistemul de fisiere


void IO_init(void)
{
	DDRA &= ~(1 << PA0);
	DDRA &= ~(1 << PA1);
    DDRA &= ~(1 << PA2);
	DDRA &= ~(1 << PA4);
	DDRA &= ~(1 << PA5);
    PORTA |= (1 << PA0) | (1 << PA1) | (1 << PA2) | (1 << PA4) | (1 << PA5);

    DDRB &= ~(1 << PB0);
    PORTB |= (1 << PB0);
}


/*---------------------------------------------------------------------------*/
/* Player audio                                                              */
/*---------------------------------------------------------------------------*/

/*
 * Four-Character Code - folosit pentru a indentifica formate de date
 */
#define FCC(c1, c2, c3, c4) \
	(((DWORD)(c4) << 24) + \
	 ((DWORD)(c3) << 16) + \
	 (( WORD)(c2) <<  8) + \
	 (( BYTE)(c1) <<  0))


uint8_t	buf[2][256];	// wave output buffers (double buffering)
const	 uint16_t	buf_size = 256;	// front and back buffer sizes
volatile uint8_t	buf_front = 0;	// front buffer index (current buffer used)
volatile uint8_t	buf_pos = 0;	// current buffer position
volatile uint8_t	buf_sync = 0;

#define BUF_FRONT	(buf[buf_front])
#define BUF_BACK	(buf[1 - buf_front])

ISR(TIMER0_COMPA_vect)
{
	OCR1B = BUF_FRONT[buf_pos++];

	// swap buffers when end is reached (end is 256 <=> overflow to 0)
	if(buf_pos == 0)
		buf_front = 1 - buf_front;
}

void timer0_start(void)
{
	// interrupt on compare A
	TIMSK0 |= (1 << OCIE0A);
	// CTC, top OCRA
	TCCR0B |= (0 << WGM02);
	TCCR0A |= (1 << WGM01) | (0 << WGM00);
	// prescaler 8
	TCCR0B |= (2 << CS00);
}

void timer0_stop(void)
{
	TCCR0B = 0;
	TCCR0A = 0;
	TIMSK0 = 0;
	OCR0A = 0;
	TCNT0 = 0;
}

void timer1_start(void)
{
	// 8-bit FastPWM
	TCCR1B |= (1 << WGM12);
	TCCR1A |= (1 << WGM10);
	// channel B inverted
	TCCR1A |= (1 << COM1B0) | (1 << COM1B1);
	// prescaler 1
	TCCR1B |= (1 << CS10);
}

void timer1_stop(void)
{
	TCCR1B = 0;
	TCCR1A = 0;
	OCR1B = 0;
	TCNT1 = 0;
}

bool continue_play()
{
	if((PINA & (1 << PA2)) == 0 && (PINA & (1 << PA6)) == 0)
		return false;

	return true;
}

/*
 * Incarca header-ul unui fisier WAVE
 *
 * @return DWORD
 * 	0 => format invalid
 * 	1 => eroare I/O
 * 	>1 => numarul de sample-uri
 */
DWORD load_header(void)
{
	DWORD size;
	WORD ret;

	// citeste header-ul (12 octeti)
	if(pf_read(BUF_FRONT, 12, &ret))
		return 1;

	if(ret != 12 || LD_DWORD(BUF_FRONT + 8) != FCC('W','A','V','E'))
		return 0;

	for(;;)
	{
		// citeste chunk ID si size
		pf_read(BUF_FRONT, 8, &ret);
		if(ret != 8)
			return 0;

		size = LD_DWORD(&BUF_FRONT[4]);

		// verifica FCC
		switch(LD_DWORD(&BUF_FRONT[0]))
		{
			// 'fmt ' chunk
			case FCC('f','m','t',' '):
				// verifica size
				if(size > 100 || size < 16) return 0;

				// citeste continutul
				pf_read(BUF_FRONT, size, &ret);
				// verifica codificarea
				if(ret != size || BUF_FRONT[0] != 1) return 0;
				// verifica numarul de canale
				if(BUF_FRONT[2] != 1 && BUF_FRONT[2] != 2) return 0;
				// verifica rezolutia
				if(BUF_FRONT[14] != 8 && BUF_FRONT[14] != 16) return 0;

				// seteaza sampling rate-ul
				OCR0A = (BYTE)(F_CPU / 8 / LD_WORD(&BUF_FRONT[4])) - 1;
				break;

			// 'data' chunk => incepe redarea
			case FCC('d','a','t','a'):
				return size;

			// 'LIST' chunk => skip
			case FCC('L','I','S','T'):
			// 'fact' chunk => skip
			case FCC('f','a','c','t'):
				pf_lseek(fs.fptr + size);
				break;

			// chunk necunoscut => eroare
			default:
				return 0;
		}
	}

	return 0;
}

/*
 * Functie care reda un fisier audio
 *
 * path - calea absoluta a fisierului
 *
 * @return UINT
 *	FR_OK daca a rulat cu succes fisierul
 */
UINT play(const char *path)
{
	FRESULT ret;

	if((ret = pf_open(path)) == FR_OK)
	{
		WORD bytes_read;

		// incarca header-ul fisierului
		DWORD current_size = load_header();
		if(current_size < buf_size)
			return FR_NO_FILE;

		// align to sector boundary
		ret = pf_lseek((fs.fptr + 511) & ~511);
		if(ret != FR_OK)
			return ret;

		// fill front buffer
		ret = pf_read(BUF_FRONT, buf_size, &bytes_read);
		if(ret != FR_OK)
			return ret;
		if(bytes_read < buf_size)
			return ret;

		// reset front buffer index
		buf_pos = 0;

		// start output
		timer0_start();
		timer1_start();
		DDRD |= (1 << PD4);

		while(continue_play())
		{
			uint8_t old_buf_front = buf_front;
			
			// fill back buffer
			ret = pf_read(BUF_BACK, buf_size, &bytes_read);
			if(ret != FR_OK)
				break;
			if(bytes_read < buf_size)
				break;

			// wait for buffer swap
			while(old_buf_front == buf_front) ;
		}

		// stop output
		DDRD &= ~(1 << PD4);
		timer1_stop();
		timer0_stop();
	}

	return ret;
}


/*---------------------------------------------------------------------------*/
/* Ceas                                                                      */
/*---------------------------------------------------------------------------*/

volatile uint8_t hours = 0;
volatile uint8_t minutes = 0;
volatile uint8_t seconds = 0;

ISR(TIMER2_COMPA_vect)
{
	static uint16_t miliseconds = 0;

	if(++miliseconds != 1000)
		return;
	miliseconds = 0;

	if(++seconds == 60)
	{
		seconds = 0;

		if(++minutes == 60)
		{
			minutes = 0;

			if(++hours == 24)
				hours = 0;
		}
	}

	char buf[9];
	snprintf(buf, sizeof(buf), "%02d:%02d:%02d", hours, minutes, seconds);

	LCD_clear_bottom_line();
	LCD_printAt(0x40, buf);
}

void timer2_init(void)
{
	// set compare at each milisecond
	OCR2A = 124;
	// interrupt on compare A
	TIMSK2 |= (1 << OCIE2A);
	// CTC, top OCRA
	TCCR2A |= (1 << WGM21);
	// prescaler 128
	TCCR2B |= (1 << CS20) | (1 << CS22);
}

char use_rainbow;

#define MUSIC "music"

/*
 * Cauta si returneaza al n-lea fisier din folder-ul primit
 */
void get_music(int n, const char *folder, char *filename)
{
	DIR dir;
	FILINFO fil;

	char i = 0;

	pf_opendir(&dir, MUSIC);
	for(i = 0; i < n; i++) {
		pf_readdir(&dir, &fil);
		if(fil.fname[0] == 0x00) {
			filename[0] = 0x00;
			return;
		}
	}

	filename[0] = 0;
	strcat(filename, folder);
	strcat(filename, "/");
	strcat(filename, fil.fname);
}

char current_file_no = 1;
char filename[32] = { 0 };
int wasThere = 0;

void showHourOnFigure()
{
    // Not implemented yet
}

ISR(PCINT1_vect)
{
    if (PINB & (1 << PB0)) {
        LCD_clear_top_line();
	    LCD_printAt(0x00, filename);
	    play(filename);
        use_rainbow = 1;
    }
}

ISR(PCINT0_vect)
{
    if (PINA & (1 << PA0)) {
        filename[0] = 0;
	    get_music(current_file_no, MUSIC, filename);
	    if (filename[0] == 0x00){
		    current_file_no = 1;
		    get_music(current_file_no, MUSIC, filename);
	    }

	    LCD_clear_top_line();	
	    LCD_printAt(0x00, filename);
	    current_file_no += 1;
    }
    if (PINA & (1 << PA2)) {
        filename[0] = 0;
	    get_music(current_file_no, MUSIC, filename);
	    if (filename[0] == 0x00){
		    current_file_no = 1;
		    get_music(current_file_no, MUSIC, filename);
	    }

	    LCD_clear_top_line();	
	    LCD_printAt(0x00, filename);
	    current_file_no -= 1;
        if (current_file_no < 1) {
            current_file_no = 1;
        }
    }
    if (PINA & (1 << PA1)) {
        minutes++;
        if (minutes >= 60) {
            minutes -= 60;
            hours++;
        }
    }
    if (PINA & (1 << PA4)) {
        hours++;
        if (hours >= 24) {
            hours -= 24;
        }
    }
    if (PINA & (1 << PA5)) {
        showHourOnFigure();
    }
}

void setHour()
{
    hours = 12;
    minutes = 00;
}

int main (void)
{
    uint8_t j = 1;
	uint8_t k = 1;

	DDRB|=_BV(ws2812_pin);

	IO_init();
	LCD_init();
	timer2_init();

    PCICR |= (1 << PCIE0) | (1 << PCIE1);
    PCMSK0 |= (1 << PCINT0) | (1 << PCINT1) | (1 << PCINT2) | (1 << PCINT4) | (1 << PCINT5);
    PCMSK1 |= (1 << PCINT8);

	sei();
    setHour();

    uint8_t i;
    for(i=MAXPIX; i>0; i--)
    {    
        led[i-1].r=0;led[i-1].g=0;led[i-1].b=0;
    }
		
    //Rainbowcolors
    colors[0].r=150; colors[0].g=150; colors[0].b=150;
    colors[1].r=255; colors[1].g=000; colors[1].b=000;//red
    colors[2].r=255; colors[2].g=100; colors[2].b=000;//orange
    colors[3].r=100; colors[3].g=255; colors[3].b=000;//yellow
    colors[4].r=000; colors[4].g=255; colors[4].b=000;//green
    colors[5].r=000; colors[5].g=100; colors[5].b=255;//light blue
    colors[6].r=000; colors[6].g=000; colors[6].b=255;//blue
    colors[7].r=100; colors[7].g=000; colors[7].b=255;//violet

    if(!wasThere)
	{
        wasThere = 1;
		LCD_clear_top_line();

		DIR dir;
		if(pf_opendir(&dir, MUSIC) != FR_OK)
		{
			LCD_printAt(0x00, MUSIC"missing");
		}
	}

	for(;;)
	{
		// mount filesystem
		LCD_printAt(0x00, "mounting...");

		if(pf_mount(&fs) != FR_OK)
		{
			// wait a while and retry
			_delay_ms(1000);
			continue;
		}

		LCD_clear_top_line();
		LCD_printAt(0x00, "mounted!");
		_delay_ms(1000);
		LCD_clear_top_line();

		for(;;)
		{
            while (use_rainbow) {
                uint8_t i=0;           
                for(i=MAXPIX; i>1; i--) 
                    led[i-1]=led[i-2];
                //change colour when colourlength is reached   
                if(k>COLORLENGTH)
                {
                    j++;
                    if(j>7)
                    {
                      j=0;
                    }

                    k=0;
                }
                k++;
                //loop colouers
                
                //fade red
                if(led[0].r<(colors[j].r-FADE))
                    led[0].r+=FADE;
                    
                if(led[0].r>(colors[j].r+FADE))
                    led[0].r-=FADE;

                if(led[0].g<(colors[j].g-FADE))
                    led[0].g+=FADE;
                    
                if(led[0].g>(colors[j].g+FADE))
                    led[0].g-=FADE;

                if(led[0].b<(colors[j].b-FADE))
                    led[0].b+=FADE;
                    
                if(led[0].b>(colors[j].b+FADE))
                    led[0].b-=FADE;

		         _delay_ms(10);
		         ws2812_sendarray((uint8_t *)led,MAXPIX*3);
            }
            _delay_ms(500);
		}
	}

	return 0;
}
